import * as THREE from 'three';
import {TextGeometry} from 'three/examples/jsm/geometries/TextGeometry';
import {FontLoader} from 'three/examples/jsm/loaders/FontLoader';

window.document.addEventListener("DOMContentLoaded", (event: Event) => new Catch_the_bus('img/Bus.jpg', 20, 5) /* Each 4 sec. within 20 sec. */);

class Catch_the_bus {
    private _time_left = this._count_down; // In sec.

    private readonly _camera = new THREE.PerspectiveCamera;
    private readonly _renderer = new THREE.WebGLRenderer;
    private readonly _scene = new THREE.Scene();

    private _animation_id: number = 0; // '0' means that animation hasn't yet started...
    private _interval_id!: number;

    // 'strictPropertyInitialization' set to 'true' requires '!':
    private _text_canvas!: HTMLCanvasElement; // DOM must be loaded...
    private _text_canvas_context!: CanvasRenderingContext2D; // DOM must be loaded...
    private _text_canvas_texture!: THREE.CanvasTexture; // DOM must be loaded...

    constructor(private readonly _image_URL: string, private readonly _count_down: number, private readonly _steps: number) {
        this._camera = new THREE.PerspectiveCamera(50, window.innerWidth / window.innerHeight, 0.1, 5000);
        this._camera.position.x += 100;
        this._camera.position.y += 100;
        this._camera.position.z = 3000;
        this._renderer = new THREE.WebGLRenderer({antialias: true});
        this._renderer.setClearColor(new THREE.Color(0xC0C0C0)/* Silver */); // For areas without filmed objects, useful for debugging!
        this._renderer.setPixelRatio((window.devicePixelRatio) ? window.devicePixelRatio : 1);
        this._renderer.setSize(window.innerWidth, window.innerHeight);

        window.document.body.appendChild(this._renderer.domElement); // DOM must be loaded...

        this._build_Axes(2000);

        const image: HTMLImageElement = new Image();
        image.onload = () => {
            this._create_3D_bus(image, this._image_URL);
            // Doesn't work because 'this' is 'undefined': 'this._animation_id = window.requestAnimationFrame(this._animate);'
            this._animation_id = window.requestAnimationFrame(this._animate.bind(this));
            // '_time_out' *ATTRIBUTE* as lambda precludes the use of 'this._time_out.bind(this)':
            this._interval_id = window.setInterval(this._time_out, this._count_down / this._steps * 1000);
        };
        image.src = this._image_URL;

        this._create_3D_panel();

        this._create_3D_text();
    }

    private _animate(timestamp?: number) {
        if (timestamp !== undefined) /* first call: 'timestamp === undefined' */
            // One may handle the elapsed time from the first call to '_animate'
            this._animation_id = window.requestAnimationFrame(this._animate.bind(this));

        const bus = this._scene.getObjectByName(this._image_URL);
        if (bus) {
            bus.rotation.y -= 0.01;
            ((bus as THREE.Mesh).material as THREE.Material).opacity -= 0.0005;
        }

        const panel = this._scene.getObjectByName(`${this._count_down}`);
        if (panel) {
            panel.rotation.x += 0.05;
            panel.rotation.y += 0.05;
            panel.rotation.z += 0.05;
        }

        this._renderer.render(this._scene, this._camera);
    }

    private _build_Axes(length: number) {
        const axes = new THREE.Object3D();
        axes.add(this._build_Axis(new THREE.Vector3(0, 0, 0), new THREE.Vector3(length, 0, 0), new THREE.Color('red'), false)); // +X
        axes.add(this._build_Axis(new THREE.Vector3(0, 0, 0), new THREE.Vector3(-length, 0, 0), new THREE.Color('red'), true)); // -X
        axes.add(this._build_Axis(new THREE.Vector3(0, 0, 0), new THREE.Vector3(0, length, 0), new THREE.Color('blue'), false)); // +Y
        axes.add(this._build_Axis(new THREE.Vector3(0, 0, 0), new THREE.Vector3(0, -length, 0), new THREE.Color('blue'), true)); // -Y
        axes.add(this._build_Axis(new THREE.Vector3(0, 0, 0), new THREE.Vector3(0, 0, length), new THREE.Color('green'), false)); // +Z
        axes.add(this._build_Axis(new THREE.Vector3(0, 0, 0), new THREE.Vector3(0, 0, -length), new THREE.Color('green'), true)); // -Z
        this._scene.add(axes);
    }

    private _build_Axis(vector3_source: THREE.Vector3, vector3_target: THREE.Vector3, color: THREE.Color, is_dashed: boolean) {
        const geometry = new THREE.BufferGeometry().setFromPoints([vector3_source, vector3_target]);
        const axis = new THREE.LineSegments(geometry, is_dashed ? new THREE.LineDashedMaterial({
            linewidth: 5,
            color: color.getHex(),
            dashSize: 5,
            gapSize: 10
        }) : new THREE.LineBasicMaterial({linewidth: 5, color: color.getHex()})); // Default 'THREE.LineBasicMaterial' is created...
        axis.computeLineDistances(); // This one is very important, otherwise dashed lines will appear as simple plain lines
        return axis;
    }

    private _create_3D_bus(image: HTMLImageElement, image_URL: string) {
        if (!(image !== undefined && image !== null && image instanceof Image && image.complete))
            throw ("Abnormal situation...");
        const canvas = window.document.createElement('canvas');
        canvas.width = image.width;
        canvas.height = image.height;
        // 'getContext' may return 'null':
        canvas.getContext('2d')!.drawImage(image, 0, 0, image.width, image.height);
        const texture = new THREE.CanvasTexture(canvas);
        texture.wrapS = THREE.RepeatWrapping;
        texture.wrapT = THREE.RepeatWrapping;
        texture.repeat.set(2, 2);
        texture.minFilter = THREE.LinearFilter;
        const bus = new THREE.Mesh(new THREE.SphereGeometry(600, 0, 0, Math.PI / 4, Math.PI / 2, Math.PI / 4, Math.PI / 2), new THREE.MeshBasicMaterial({
            map: texture,
            opacity: 1,
            side: THREE.DoubleSide,
            transparent: true
        }));
        bus.name = image_URL;
        this._scene.add(bus);
    }

    private _create_3D_bus_(texture: THREE.CanvasTexture, image_URL: string) {
        texture.wrapS = THREE.RepeatWrapping;
        texture.wrapT = THREE.RepeatWrapping;
        texture.repeat.set(2, 2);
        texture.minFilter = THREE.LinearFilter;
        const bus = new THREE.Mesh(new THREE.SphereGeometry(600, 0, 0, Math.PI / 4, Math.PI / 2, Math.PI / 4, Math.PI / 2), new THREE.MeshStandardMaterial({
            map: texture,
            opacity: 1,
            side: THREE.DoubleSide,
            transparent: true
        }));
        bus.name = image_URL;
        this._scene.add(bus);
    }

    private _create_3D_panel() {
        this._text_canvas = window.document.createElement('canvas');
        this._text_canvas.width = 400;
        this._text_canvas.height = 300;
        // Cast tells TypeScript that 'getContext' won't return 'null', but not fully reliable:
        this._text_canvas_context = <CanvasRenderingContext2D>this._text_canvas.getContext('2d');
        // Background:
        this._text_canvas_context.beginPath();
        this._text_canvas_context.rect(0, 0, this._text_canvas.width, this._text_canvas.height);
        this._text_canvas_context.fillStyle = "silver";
        this._text_canvas_context.fill();

        this._text_canvas_context.fillStyle = "gray";
        this._text_canvas_context.font = "Bold 40px Arial";
        this._text_canvas_context.textAlign = "center";
        this._text_canvas_context.textBaseline = "middle";
        this._text_canvas_context.fillText(this._time_left + " sec. left...", this._text_canvas.width / 2, this._text_canvas.height / 2);
        this._text_canvas_texture = new THREE.CanvasTexture(this._text_canvas);

        let panel = new THREE.Mesh(new THREE.PlaneGeometry(500, 500), new THREE.MeshBasicMaterial({
//            color: new THREE.Color(0xFFFF00),
            map: this._text_canvas_texture,
//            side: THREE.DoubleSide
        }));
        panel.name = "PANEL";
        panel.translateY(-this._time_left * 50); // Down
        panel.quaternion.setFromAxisAngle(new THREE.Vector3(1, 0, 0), -Math.PI / 8);
        this._scene.add(panel);
    }

    private _create_3D_text() {
        new Promise<TextGeometry>(get => {
            new FontLoader().load('./fonts/helvetiker_regular.typeface.json', font => {
                get(new TextGeometry(`${this._count_down}` + " sec. to catch the bus...", {
                    font: font,
                    size: 5,
                    height: 2,
                    curveSegments: 20,
                    bevelEnabled: false, // "biseau" in French!
                    bevelThickness: 5,
                    bevelSize: 5,
                    bevelSegments: 5
                }));
            });
        }).then((text_geometry: TextGeometry) => {
            text_geometry.computeBoundingBox();
            text_geometry.center();
            const text = new THREE.Mesh(text_geometry, new THREE.MeshBasicMaterial({
                color: new THREE.Color(0xFF0000) // Red
            }));
            text.name = `${this._count_down}`;
            text.scale.multiplyScalar(25);
            this._scene.add(text);
        });
    }

    // '_time_out' *ATTRIBUTE* as lambda:
    private _time_out: () => void = () => {
        // Unreliable, we just hope that the context is already acquired:
        window.console.assert(this._text_canvas_context);
        // this._text_canvas_context.clearRect(0, 0, this._text_canvas.width, this._text_canvas.height);
        // Background:
        this._text_canvas_context.beginPath();
        this._text_canvas_context.rect(0, 0, this._text_canvas.width, this._text_canvas.height);
        this._text_canvas_context.fillStyle = "silver";
        this._text_canvas_context.fill();
        if ((this._time_left -= this._count_down / this._steps) > 0) {
            this._scene.getObjectByName("PANEL")!.translateY(this._time_left * 50);
            this._text_canvas_context.fillStyle = 'gray';
            this._text_canvas_context.fillText(this._time_left + " sec. left...", this._text_canvas.width / 2, this._text_canvas.height / 2);
            this._text_canvas_texture.needsUpdate = true; // Re-send texture to WebGL...
        } else {
            window.clearInterval(this._interval_id); // Stop timer...
            ((this._scene.getObjectByName(this._image_URL) as THREE.Mesh).material as THREE.MeshStandardMaterial).color.setHex(0xFF0000);
            ((this._scene.getObjectByName(this._image_URL) as THREE.Mesh).material as THREE.Material).opacity = 0.5;
            this._text_canvas_context.fillStyle = "#FF0000";
            this._text_canvas_context.fillText("The end...", this._text_canvas.width / 2, this._text_canvas.height / 2);
            this._text_canvas_texture.needsUpdate = true; // Re-send texture to WebGL...
            this._renderer.render(this._scene, this._camera); // Very last rendering...
            window.cancelAnimationFrame(this._animation_id);
        }
    }
}
